import numpy as np
from numpy.linalg import lstsq, norm
from scipy.special import expit
from sklearn.cluster import MiniBatchKMeans


def one_hot_encode(y, n_classes):
    n_obs = len(y)
    res = np.zeros((n_obs, n_classes))
    for i in range(n_obs):
        res[i, y[i]] = 1
    return res


def update_proba(
    probs, misclass, w_new, w_prev, v_prev, eta, gamma, n_obs, n_classes
):
    new_probs = probs.copy()
    v_new = 0
    for idx in range(n_obs):
        if misclass[idx] != 1000:
            diff_proba = max(w_new[idx], 2.220446049250313e-16) - w_prev[idx]
            v_new = diff_proba
            new_probs[idx, misclass[idx]] -= (v_prev * gamma - v_new) / eta
    return new_probs, v_new


def fit_adaopt(
    X,
    y,
    n_iterations,
    n_classes,
    learning_rate,
    reg_lambda,
    reg_alpha,
    eta,
    gamma,
    tolerance,
):
    scaled_X = X / norm(X, ord=2, axis=1)[:, None]
    Y = one_hot_encode(y, n_classes)
    beta = lstsq(scaled_X, Y, rcond=None)[0]
    probs = expit(np.dot(scaled_X, beta))
    probs /= np.sum(probs, axis=1)[:, None]
    preds = np.argmax(probs, axis=1)
    w_prev = np.repeat(1.0 / len(X), len(X))
    w_new = np.repeat(0.0, len(X))
    misclass = np.repeat(1000, len(X))
    v_prev = 0
    alphas = np.zeros(n_iterations)
    err_bound = 1.0 - 1.0 / n_classes

    for m in range(n_iterations):
        for i in range(len(X)):
            if y[i] != preds[i]:
                misclass[i] = preds[i]
            else:
                misclass[i] = 1000

        err_m = np.sum(misclass != 1000) / len(X)
        err_m += reg_lambda * (
            reg_alpha * np.sum(np.abs(w_prev))
            + (1 - reg_alpha) * 0.5 * np.sum(np.power(w_prev, 2))
        )
        err_m = min(max(err_m, 2.220446049250313e-16), err_bound)

        alpha_m = learning_rate * np.log((n_classes - 1) * (1 / err_m - 1))
        alphas[m] = alpha_m
        w_prev *= np.exp(alpha_m * (misclass != 1000))
        w_prev /= np.sum(w_prev)
        w_new = w_prev

        probs, v_prev = update_proba(
            probs,
            misclass,
            w_new,
            w_prev,
            v_prev,
            eta,
            gamma,
            len(X),
            n_classes,
        )
        preds = np.argmax(probs, axis=1)

        if np.abs(np.max(alphas)) - np.abs(np.min(alphas)) <= tolerance:
            n_iterations = m
            break

    return {"probs": probs, "alphas": alphas, "n_iterations": n_iterations}


def predict_proba_adaopt(
    X_test,
    scaled_X_train,
    probs_train,
    k,
    n_clusters,
    seed,
    batch_size=100,
    type_dist="euclidean",
    cache=True,
):
    n_test, n_train = len(X_test), len(scaled_X_train)
    n_classes = probs_train.shape[1]
    out_probs = np.zeros((n_test, n_classes), dtype=np.double)
    probs_train_ = probs_train

    if n_clusters > 0:
        probs_train_ = np.zeros((n_clusters, n_classes), dtype=np.double)
        kmeans = MiniBatchKMeans(
            n_clusters=n_clusters, batch_size=batch_size, random_state=seed
        ).fit(scaled_X_train)
        scaled_X_train = kmeans.cluster_centers_

        for m in range(n_clusters):
            index_train = np.where(kmeans.labels_ == m)[0]
            avg_probs = np.average(probs_train[index_train], axis=0)
            probs_train_[m] = avg_probs

    dist_mat = (
        np.sum(X_test**2, axis=1)[:, None]
        + np.sum(scaled_X_train**2, axis=1)
        - 2 * np.dot(X_test, scaled_X_train.T)
    )

    for i in range(n_test):
        dists_test_i = dist_mat[i]
        kmin_test_i = np.argsort(dists_test_i)[:k]
        weights_test_i = 1 / np.maximum(
            dists_test_i[kmin_test_i], np.finfo(float).eps
        )
        weights_test_i /= np.sum(weights_test_i)
        probs_test_i = probs_train_[kmin_test_i]
        avg_probs_i = np.sum(probs_test_i * weights_test_i[:, None], axis=0)
        out_probs[i] = avg_probs_i

    out_probs_ = expit(out_probs)
    out_probs_ /= np.sum(out_probs_, axis=1)[:, None]
    return out_probs_


def predict_adaopt(
    X_test,
    scaled_X_train,
    probs_train,
    k,
    n_clusters,
    seed,
    batch_size=100,
    type_dist="euclidean",
    cache=True,
):
    return np.argmax(
        predict_proba_adaopt(
            X_test,
            scaled_X_train,
            probs_train,
            k,
            n_clusters,
            seed,
            batch_size,
            type_dist,
            cache,
        ),
        axis=1,
    )



import numpy as np
import pickle
from joblib import Parallel, delayed
from joblib import wrap_non_picklable_objects
from sklearn.base import BaseEstimator
from sklearn.base import ClassifierMixin
from numpy.linalg import norm
from tqdm import tqdm
from ..utils import subsample
from ..utils import cluster

try:
    from . import _adaoptc as adaoptc
except ImportError:
    import _adaoptc as adaoptc


class AdaOpt(BaseEstimator, ClassifierMixin):
    """AdaOpt classifier.

    Attributes:

        n_iterations: int
            number of iterations of the optimizer at training time.

        learning_rate: float
            controls the speed of the optimizer at training time.

        reg_lambda: float
            L2 regularization parameter for successive errors in the optimizer
            (at training time).

        reg_alpha: float
            L1 regularization parameter for successive errors in the optimizer
            (at training time).

        eta: float
            controls the slope in gradient descent (at training time).

        gamma: float
            controls the step size in gradient descent (at training time).

        k: int
            number of nearest neighbors selected at test time for classification.

        tolerance: float
            controls early stopping in gradient descent (at training time).

        n_clusters: int
            number of clusters, if MiniBatch k-means is used at test time
            (for faster prediction).

        batch_size: int
            size of the batch, if MiniBatch k-means is used at test time
            (for faster prediction).

        row_sample: float
            percentage of rows chosen from training set (by stratified subsampling,
            for faster prediction).

        type_dist: str
            distance used for finding the nearest neighbors; currently `euclidean-f`
            (euclidean distances calculated as whole), `euclidean` (euclidean distances
            calculated row by row), `cosine` (cosine distance).

        n_jobs: int
            number of cpus for parallel processing (default: None)

        verbose: int
            progress bar for parallel processing (yes = 1) or not (no = 0)

        cache: boolean
            if the nearest neighbors are cached or not, for faster retrieval in
            subsequent calls.

        n_clusters_input: int
            number of clusters (a priori) for clustering the features

        clustering_method: str
            clustering method: currently 'kmeans', 'gmm'

        cluster_scaling: str
            scaling method for clustering: currently 'standard', 'robust', 'minmax'

        seed: int
            reproducibility seed for nodes_sim=='uniform', clustering and dropout.

    """

    def __init__(
        self,
        n_iterations=50,
        learning_rate=0.3,
        reg_lambda=0.1,
        reg_alpha=0.5,
        eta=0.01,
        gamma=0.01,
        k=3,
        tolerance=0,
        n_clusters=0,
        batch_size=100,
        row_sample=0.8,
        type_dist="euclidean-f",
        n_jobs=None,
        verbose=0,
        cache=True,
        n_clusters_input=0,
        clustering_method="kmeans",
        cluster_scaling="standard",
        seed=123,
    ):
        if n_clusters_input > 0:
            assert clustering_method in (
                "kmeans",
                "gmm",
            ), "`clustering_method` must be in ('kmeans', 'gmm')"
            assert cluster_scaling in (
                "standard",
                "robust",
                "minmax",
            ), "`cluster_scaling` must be in ('standard', 'robust', 'minmax')"

        assert type_dist in (
            "euclidean",
            "manhattan",
            "euclidean-f",
            "cosine",
        ), "must have: `type_dist` in ('euclidean', 'manhattan', 'euclidean-f', 'cosine') "

        self.n_iterations = n_iterations
        self.learning_rate = learning_rate
        self.reg_lambda = reg_lambda
        self.reg_alpha = reg_alpha
        self.eta = eta
        self.gamma = gamma
        self.k = k
        self.tolerance = tolerance
        self.n_clusters = n_clusters
        self.batch_size = batch_size
        self.row_sample = row_sample
        self.type_dist = type_dist
        self.n_jobs = n_jobs
        self.cache = cache
        self.verbose = verbose
        self.n_clusters_input = n_clusters_input
        self.clustering_method = clustering_method
        self.cluster_scaling = cluster_scaling
        self.scaler_, self.label_encoder_, self.clusterer_ = None, None, None
        self.seed = seed

    def fit(self, X, y, **kwargs):
        """Fit AdaOpt to training data (X, y)

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            y: array-like, shape = [n_samples]
                Target values.

            **kwargs: additional parameters to be passed to self.cook_training_set.

        Returns:

            self: object.

        """

        if self.n_clusters_input > 0:
            clustered_X, self.scaler_, self.label_encoder_, self.clusterer_ = (
                cluster(
                    X,
                    n_clusters=self.n_clusters_input,
                    method=self.clustering_method,
                    type_scaling=self.cluster_scaling,
                    training=True,
                    seed=self.seed,
                )
            )
            X = np.column_stack((X.copy(), clustered_X))

        if self.row_sample < 1:
            index_subsample = subsample(
                y, row_sample=self.row_sample, seed=self.seed
            )
            y_ = y[index_subsample]
            X_ = X[index_subsample, :]
        else:
            y_ = pickle.loads(pickle.dumps(y, -1))
            X_ = pickle.loads(pickle.dumps(X, -1))

        n, p = X_.shape

        n_classes = len(np.unique(y_))

        assert n == len(y_), "must have X.shape[0] == len(y)"

        res = adaoptc.fit_adaopt(
            X=np.asarray(X_).astype(np.float64),
            y=np.asarray(y_).astype(np.int64),
            n_iterations=self.n_iterations,
            n_X=n,
            p_X=p,
            n_classes=n_classes,
            learning_rate=self.learning_rate,
            reg_lambda=self.reg_lambda,
            reg_alpha=self.reg_alpha,
            eta=self.eta,
            gamma=self.gamma,
            tolerance=self.tolerance,
        )

        self.probs_training = res["probs"]
        self.training_accuracy = res["training_accuracy"]
        self.alphas = res["alphas"]
        self.n_iterations = res["n_iterations"]
        self.scaled_X_train = np.array(res["scaled_X_train"], dtype=np.float64)
        self.n_classes_ = len(np.unique(y))  # for compatibility with sklearn
        return self

    def predict(self, X, **kwargs):
        """Predict test data X.

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            **kwargs: additional parameters to be passed to `predict_proba`

        Returns:

            model predictions: {array-like}

        """

        return np.argmax(self.predict_proba(X, **kwargs), axis=1)

    def predict_proba(self, X, **kwargs):
        """Predict probabilities for test data X.

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            **kwargs: additional parameters to be passed to
                self.cook_test_set

        Returns:

            probability estimates for test data: {array-like}

        """

        n_train, p_train = self.scaled_X_train.shape

        if self.n_clusters_input > 0:
            X = np.column_stack(
                (
                    X.copy(),
                    cluster(
                        X,
                        training=False,
                        scaler=self.scaler_,
                        label_encoder=self.label_encoder_,
                        clusterer=self.clusterer_,
                        seed=self.seed,
                    ),
                )
            )

        n_test = X.shape[0]

        if self.n_jobs is None:
            return adaoptc.predict_proba_adaopt(
                X_test=np.asarray(X, order="C").astype(np.float64),
                scaled_X_train=np.asarray(
                    self.scaled_X_train, order="C"
                ).astype(np.float64),
                n_test=n_test,
                n_train=n_train,
                probs_train=self.probs_training,
                k=self.k,
                n_clusters=self.n_clusters,
                batch_size=self.batch_size,
                type_dist=self.type_dist,
                cache=self.cache,
                seed=self.seed,
            )

        # parallel: self.n_jobs is not None
        assert self.type_dist in (
            "euclidean",
            "manhattan",
            "cosine",
        ), "must have: `self.type_dist` in ('euclidean', 'manhattan', 'cosine') "

        scaled_X_test = X / norm(X, ord=2, axis=1)[:, None]

        if self.type_dist == "euclidean":

            @delayed
            @wrap_non_picklable_objects
            def multiproc_func(i):
                dists_test_i = adaoptc.distance_to_mat_euclidean2(
                    np.asarray(scaled_X_test.astype(np.float64), order="C")[
                        i, :
                    ],
                    np.asarray(
                        self.scaled_X_train.astype(np.float64), order="C"
                    ),
                    np.zeros(n_train),
                    n_train,
                    p_train,
                )

                kmin_test_i = adaoptc.find_kmin_x(
                    dists_test_i, n_x=n_train, k=self.k, cache=self.cache
                )

                weights_test_i = adaoptc.calculate_weights(kmin_test_i[0])

                probs_test_i = adaoptc.calculate_probs(
                    kmin_test_i[1], self.probs_training
                )

                return adaoptc.average_probs(
                    probs=probs_test_i, weights=weights_test_i
                )

        if self.type_dist == "manhattan":

            @delayed
            @wrap_non_picklable_objects
            def multiproc_func(i):
                dists_test_i = adaoptc.distance_to_mat_manhattan2(
                    np.asarray(scaled_X_test.astype(np.float64), order="C")[
                        i, :
                    ],
                    np.asarray(
                        self.scaled_X_train.astype(np.float64), order="C"
                    ),
                    np.zeros(n_train),
                    n_train,
                    p_train,
                )

                kmin_test_i = adaoptc.find_kmin_x(
                    dists_test_i, n_x=n_train, k=self.k, cache=self.cache
                )

                weights_test_i = adaoptc.calculate_weights(kmin_test_i[0])

                probs_test_i = adaoptc.calculate_probs(
                    kmin_test_i[1], self.probs_training
                )

                return adaoptc.average_probs(
                    probs=probs_test_i, weights=weights_test_i
                )

        if self.type_dist == "cosine":

            @delayed
            @wrap_non_picklable_objects
            def multiproc_func(i, *args):
                dists_test_i = adaoptc.distance_to_mat_cosine2(
                    np.asarray(scaled_X_test.astype(np.float64), order="C")[
                        i, :
                    ],
                    np.asarray(
                        self.scaled_X_train.astype(np.float64), order="C"
                    ),
                    np.zeros(n_train),
                    n_train,
                    p_train,
                )

                kmin_test_i = adaoptc.find_kmin_x(
                    dists_test_i, n_x=n_train, k=self.k, cache=self.cache
                )

                weights_test_i = adaoptc.calculate_weights(kmin_test_i[0])

                probs_test_i = adaoptc.calculate_probs(
                    kmin_test_i[1], self.probs_training
                )

                return adaoptc.average_probs(
                    probs=probs_test_i, weights=weights_test_i
                )

        if self.verbose == 1:
            res = Parallel(n_jobs=self.n_jobs, prefer="threads")(
                (multiproc_func)(m) for m in tqdm(range(n_test))
            )

        else:
            res = Parallel(n_jobs=self.n_jobs, prefer="threads")(
                (multiproc_func)(m) for m in range(n_test)
            )

        return np.asarray(res)



import sys
import logging
import os

from ._config import get_config, set_config, config_context

logger = logging.getLogger(__name__)


# PEP0440 compatible formatted version, see:
# https://www.python.org/dev/peps/pep-0440/
#
# Generic release markers:
#   X.Y
#   X.Y.Z   # For bugfix releases
#
# Admissible pre-release markers:
#   X.YaN   # Alpha release
#   X.YbN   # Beta release
#   X.YrcN  # Release Candidate
#   X.Y     # Final release
#
# Dev branch marker is: 'X.Y.dev' or 'X.Y.devN' where N is an integer.
# 'X.Y.dev0' is the canonical version of 'X.Y.dev'
#
# __version__ = "0.10.0"


# On OSX, we can get a runtime error due to multiple OpenMP libraries loaded
# simultaneously. This can happen for instance when calling BLAS inside a
# prange. Setting the following environment variable allows multiple OpenMP
# libraries to be loaded. It should not degrade performances since we manually
# take care of potential over-subcription performance issues, in sections of
# the code where nested OpenMP loops can happen, by dynamically reconfiguring
# the inner OpenMP runtime to temporarily disable it while under the scope of
# the outer OpenMP parallel section.
os.environ.setdefault("KMP_DUPLICATE_LIB_OK", "True")

# Workaround issue discovered in intel-openmp 2019.5:
# https://github.com/ContinuumIO/anaconda-issues/issues/11294
os.environ.setdefault("KMP_INIT_AT_FORK", "FALSE")

try:
    # This variable is injected in the __builtins__ by the build
    # process. It is used to enable importing subpackages of mlsauce when
    # the binaries are not built
    # mypy error: Cannot determine type of '__MLSAUCE_SETUP__'
    __MLSAUCE_SETUP__  # type: ignore
except NameError:
    __MLSAUCE_SETUP__ = False

if __MLSAUCE_SETUP__:
    sys.stderr.write("Partial import of mlsauce during the build process.\n")
    # We are not importing the rest of scikit-learn during the build
    # process, as it may not be compiled yet

else:
    from .adaopt import AdaOpt
    from .booster import LSBoostClassifier, LSBoostRegressor
    from .datasets import download
    from .elasticnet import ElasticNetRegressor
    from .lasso import LassoRegressor
    from .ridge import RidgeRegressor
    from .stump import StumpClassifier

    # from .encoders import corrtarget_encoder

    __all__ = [
        "AdaOpt",
        "LSBoostClassifier",
        "StumpClassifier",
        "ElasticNetRegressor",
        "LassoRegressor",
        "LSBoostRegressor",
        "RidgeRegressor",
        # Other imports
        # "corrtarget_encoder",
        "download",
        # Non-modules:
        "get_config",
        "set_config",
        "config_context",
    ]


def setup_module(module):
    """Fixture for the tests to assure globally controllable seeding of RNGs"""
    import os
    import numpy as np
    import random

    # Check if a random seed exists in the environment, if not create one.
    _random_seed = os.environ.get("MLSAUCE_SEED", None)
    if _random_seed is None:
        _random_seed = np.random.uniform() * np.iinfo(np.int32).max
    _random_seed = int(_random_seed)
    print("I: Seeding RNGs with %r" % _random_seed)
    np.random.seed(_random_seed)
    random.seed(_random_seed)
